searchNodes=[{"ref":"Parent.html","title":"Parent","type":"module","doc":"Functions for implementing a parent process. A parent process has the following properties: It traps exits. It tracks its children inside the process dictionary. Before terminating, it stops its children synchronously, in the reverse startup order. In most cases the simplest option is to start a parent process using a higher-level abstraction such as Parent.GenServer. In this case you will use a subset of the API from this module to start, stop, and enumerate your children. If available parent behaviours don&#39;t fit your purposes, you can consider building your own behaviour or a concrete process. In this case, the functions of this module will provide the necessary plumbing. To implement a parent process you need to do the following: Invoke initialize/0 when the process is started. Use functions such as start_child/1 to work with child processes. When a message is received, invoke handle_message/1 before handling the message yourself. If you receive a shutdown exit message from your parent, stop the process. Before terminating, invoke shutdown_all/1 to stop all the children. Use :infinity as the shutdown strategy for the parent process, and :supervisor for its type. If the process is a GenServer, handle supervisor calls (see supervisor_which_children/0 and supervisor_count_children/0). Implement format_status/2 (see Parent.GenServer for details) where applicable. If the parent process is powered by a non-interactive code (e.g. Task), make sure to receive messages sent to that process, and handle them properly (see points 3 and 4). You can take a look at the code of Parent.GenServer for specific details."},{"ref":"Parent.html#await_child_termination/2","title":"Parent.await_child_termination/2","type":"function","doc":"Awaits for the child to terminate. If the function succeeds, handle_child_terminated/5 will not be invoked."},{"ref":"Parent.html#child?/1","title":"Parent.child?/1","type":"function","doc":"Returns true if the child process is still running, false otherwise. Note that this function might return true even if the child has terminated. This can happen if the corresponding :EXIT message still hasn&#39;t been processed."},{"ref":"Parent.html#child_id/1","title":"Parent.child_id/1","type":"function","doc":"Returns the id of a child process with the given pid."},{"ref":"Parent.html#child_meta/1","title":"Parent.child_meta/1","type":"function","doc":"Returns the meta associated with the given child id."},{"ref":"Parent.html#child_pid/1","title":"Parent.child_pid/1","type":"function","doc":"Returns the pid of a child process with the given id."},{"ref":"Parent.html#children/0","title":"Parent.children/0","type":"function","doc":"Returns the list of running child processes."},{"ref":"Parent.html#handle_message/1","title":"Parent.handle_message/1","type":"function","doc":"Should be invoked by the parent process for each incoming message. If the given message is not handled, this function returns nil. In such cases, the client code should perform standard message handling. Otherwise, the message has been handled by the parent, and the client code doesn&#39;t shouldn&#39;t treat this message as a standard message (e.g. by calling handle_info of the callback module). However, in some cases, a client might want to do some special processing, so the return value will contain information which might be of interest to the client. Possible values are: {:EXIT, pid, id, child_meta, reason :: term} - a child process has terminated :ignore - Parent handled this message, but there&#39;s no useful information to return Note that you don&#39;t need to invoke this function in a Parent.GenServer callback module."},{"ref":"Parent.html#initialize/0","title":"Parent.initialize/0","type":"function","doc":"Initializes the state of the parent process. This function should be invoked once inside the parent process before other functions from this module are used. If a parent behaviour, such as Parent.GenServer, is used, this function must not be invoked."},{"ref":"Parent.html#initialized?/0","title":"Parent.initialized?/0","type":"function","doc":"Returns true if the parent state is initialized."},{"ref":"Parent.html#num_children/0","title":"Parent.num_children/0","type":"function","doc":"Returns the count of running child processes."},{"ref":"Parent.html#shutdown_all/1","title":"Parent.shutdown_all/1","type":"function","doc":"Terminates all running child processes. Children are terminated synchronously, in the reverse order from the order they have been started in."},{"ref":"Parent.html#shutdown_child/1","title":"Parent.shutdown_child/1","type":"function","doc":"Terminates the child. This function waits for the child to terminate. In the case of explicit termination, handle_child_terminated/5 will not be invoked."},{"ref":"Parent.html#start_child/1","title":"Parent.start_child/1","type":"function","doc":"Starts the child described by the specification."},{"ref":"Parent.html#supervisor_count_children/0","title":"Parent.supervisor_count_children/0","type":"function","doc":"Should be invoked by the behaviour when handling :count_children GenServer call. See supervisor_which_children/0 for details."},{"ref":"Parent.html#supervisor_which_children/0","title":"Parent.supervisor_which_children/0","type":"function","doc":"Should be invoked by the behaviour when handling :which_children GenServer call. You only need to invoke this function if you&#39;re implementing a parent process using a behaviour which forwards GenServer call messages to the handle_call callback. In such cases you need to respond to the client with the result of this function. Note that parent behaviours such as Parent.GenServer will do this automatically. If no translation of GenServer messages is taking place, i.e. if you&#39;re handling all messages in their original shape, this function will be invoked through handle_message/1."},{"ref":"Parent.html#update_child_meta/2","title":"Parent.update_child_meta/2","type":"function","doc":"Updates the meta of the given child process."},{"ref":"Parent.html#t:child/0","title":"Parent.child/0","type":"type","doc":""},{"ref":"Parent.html#t:child_id/0","title":"Parent.child_id/0","type":"type","doc":""},{"ref":"Parent.html#t:child_meta/0","title":"Parent.child_meta/0","type":"type","doc":""},{"ref":"Parent.html#t:child_spec/0","title":"Parent.child_spec/0","type":"type","doc":""},{"ref":"Parent.html#t:handle_message_response/0","title":"Parent.handle_message_response/0","type":"type","doc":""},{"ref":"Parent.html#t:shutdown/0","title":"Parent.shutdown/0","type":"type","doc":""},{"ref":"Parent.html#t:start/0","title":"Parent.start/0","type":"type","doc":""},{"ref":"Parent.GenServer.html","title":"Parent.GenServer","type":"behaviour","doc":"A GenServer extension which simplifies parenting of child processes. This behaviour helps implementing a GenServer which also needs to directly start child processes and handle their termination. Starting the process The usage is similar to GenServer. You need to use the module and start the process: def MyParentProcess do use Parent.GenServer def start_link(arg) do Parent.start_link(__MODULE__, arg, options \\\\ []) end end The expression use Parent.GenServer will also inject use GenServer into your code. Your parent process is a GenServer, and this behaviour doesn&#39;t try to hide it. Except when starting the process, you work with the parent exactly as you work with any GenServer, using the same functions, and writing the same callbacks: def MyParentProcess do use Parent.GenServer def do_something(pid, arg), do: GenServer.call(pid, {:do_something, arg}) ... @impl GenServer def init(arg), do: {:ok, initial_state(arg)} @impl GenServer def handle_call({:do_something, arg}, _from, state), do: {:reply, response(state, arg), next_state(state, arg)} end Compared to plain GenServer, there are following differences: A Parent.GenServer traps exits by default. The generated child_spec/1 has the :shutdown configured to :infinity. The generated child_spec/1 specifies the :type configured to :supervisor Starting child processes To start a child process, you can invoke Parent.start_child/1 in the parent process: def handle_call(...) do Parent.start_child(child_spec) ... end The function takes a child spec map which is similar to Supervisor child specs. The map has the following keys: :id (required) - a term uniquely identifying the child :start (required) - an MFA, or a zero arity lambda invoked to start the child :meta (optional) - a term associated with the started child, defaults to nil :shutdown (optional) - same as with Supervisor, defaults to 5000 :timeout (optional) - timeout after which the child is killed by the parent, see the timeout section below, defaults to :infinity The function described with :start needs to start a linked process and return the result as {:ok, pid}. For example: Parent.start_child(%{ id: :hello_world, start: {Task, :start_link, [fn -&gt; IO.puts &quot;Hello, World!&quot; end]} }) You can also pass a zero-arity lambda for :start: Parent.start_child(%{ id: :hello_world, start: fn -&gt; Task.start_link(fn -&gt; IO.puts &quot;Hello, World!&quot; end) end }) Finally, a child spec can also be a module, or a {module, arg} function. This works similarly to supervisor specs, invoking module.child_spec/1 is which must provide the final child specification. Handling child termination When a child process terminates, handle_child_terminated/5 will be invoked. The default implementation is injected into the module, but you can of course override it: @impl Parent.GenServer def handle_child_terminated(id, child_meta, pid, reason, state) do ... {:noreply, state} end The return value of handle_child_terminated is the same as for handle_info. Timeout If a positive integer is provided via the :timeout option, the parent will terminate the child if it doesn&#39;t stop within the given time. In this case, handle_child_terminated/5 will be invoked with the exit reason :timeout. Working with child processes The Parent module provides various functions for managing child processes. For example, you can enumerate running children with Parent.children/0, fetch child meta with Parent.child_meta/1, or terminate a child process with Parent.shutdown_child/1. Termination The behaviour takes down the child processes during termination, to ensure that no child process is running after the parent has terminated. The children are terminated synchronously, one by one, in the reverse start order. The termination of the children is done after the terminate/1 callback returns. Therefore in terminate/1 the child processes are still running, and you can interact with them. Supervisor compliance A process powered by Parent.GenServer can handle supervisor specific messages, which means that for all intents and purposes, such process is treated as a supervisor. As a result, children of parent will be included in the hot code reload process."},{"ref":"Parent.GenServer.html#await_child_termination/2","title":"Parent.GenServer.await_child_termination/2","type":"function","doc":"See Parent.await_child_termination/2."},{"ref":"Parent.GenServer.html#child?/1","title":"Parent.GenServer.child?/1","type":"function","doc":"See Parent.child?/1."},{"ref":"Parent.GenServer.html#child_id/1","title":"Parent.GenServer.child_id/1","type":"function","doc":"See Parent.child_id/1."},{"ref":"Parent.GenServer.html#child_meta/1","title":"Parent.GenServer.child_meta/1","type":"function","doc":"See Parent.child_meta/1."},{"ref":"Parent.GenServer.html#child_pid/1","title":"Parent.GenServer.child_pid/1","type":"function","doc":"See Parent.child_pid/1."},{"ref":"Parent.GenServer.html#children/0","title":"Parent.GenServer.children/0","type":"function","doc":"See Parent.children/0."},{"ref":"Parent.GenServer.html#c:handle_child_terminated/5","title":"Parent.GenServer.handle_child_terminated/5","type":"callback","doc":"Invoked when a child has terminated."},{"ref":"Parent.GenServer.html#num_children/0","title":"Parent.GenServer.num_children/0","type":"function","doc":"See Parent.num_children/0."},{"ref":"Parent.GenServer.html#shutdown_all/1","title":"Parent.GenServer.shutdown_all/1","type":"function","doc":"See Parent.shutdown_all/1."},{"ref":"Parent.GenServer.html#shutdown_child/1","title":"Parent.GenServer.shutdown_child/1","type":"function","doc":"See Parent.shutdown_child/1."},{"ref":"Parent.GenServer.html#start_child/1","title":"Parent.GenServer.start_child/1","type":"function","doc":"See Parent.start_child/1."},{"ref":"Parent.GenServer.html#start_link/3","title":"Parent.GenServer.start_link/3","type":"function","doc":"Starts the parent process."},{"ref":"Parent.GenServer.html#update_child_meta/2","title":"Parent.GenServer.update_child_meta/2","type":"function","doc":"See Parent.update_child_meta/2."},{"ref":"Parent.GenServer.html#t:state/0","title":"Parent.GenServer.state/0","type":"type","doc":""},{"ref":"Periodic.html","title":"Periodic","type":"module","doc":"Periodic job execution. Quick start It is recommended (but not required) to implement the job in a dedicated module. For example: defmodule SomeCleanup do def child_spec(_arg) do Periodic.child_spec( id: __MODULE__, run: &amp;run/0, every: :timer.hours(1) ) end defp run(), do: # ... end With such module implemented, you can place the job somewhere in the supervision tree: Supervisor.start_link( [ SomeCleanup, # ... ], # ... ) You can of course start multiple periodic jobs in the system, and they don&#39;t have to be the children of the same supervisor. You&#39;re advised to place the job in the proper part of the supervision tree. For example, a database cleanup job should share the ancestor with the repo, while a job working with Phoenix channels should share the ancestor with the endpoint. As mentioned, you don&#39;t need to create a dedicated module to run a job. It&#39;s also possible to provide {Periodic, opts} in the supervisor child list. Finally, if you need more runtime flexibility, you can also start the job with start_link/1. Process structure The process started with start_link is called the scheduler. This is the process which regularly &quot;ticks&quot; in the given interval and executes the job. The job is executed in a separate one-off process, which is the child of the scheduler. When the job is done, the job process stops. Therefore, each job instance is running in a separate process. Depending on the overlapping mode (see the :on_overlap option), it can happen that multiple instances of the same job are running simultaneously. Options :run (required) - Zero arity function or MFA invoked to run the job. This function is invoked in a separate one-off process which is a child of the scheduler. :every (required) - Time in milliseconds between two consecutive job executions (see :delay_mode option for details). :initial_delay - Time in milliseconds before the first execution of the job. If not provided, the default value of :every is used. In other words, the first execution will by default take place after the :initial_delay interval has passed. :delay_mode - Controls how the :every interval is interpreted. Following options are possible: :regular (default) - :every represents the time between two consecutive starts :shifted - :every represents the time between the termination of the previous and the start of the next instance. See the &quot;Delay mode&quot; section for more details. :when - Function which acts as an additional runtime guard to decide if the job will be started. This can be useful for implementing fixed scheduled jobs. See the &quot;Fixed scheduling&quot; section for details. :on_overlap - Defines the desired behaviour when the job is about to be started while the previous instance is still running. :run (default) - always start the new job :ignore - don&#39;t start the new job if the previous instance is still running :stop_previous - stop the previous instance before starting the new one :timeout - Defines the maximum running time of the job. If the job doesn&#39;t finish in the given time, it is forcefully terminated. In this case, the job&#39;s shutdown specification is ignored. Defaults to :infinity :job_shutdown - Shutdown value of the job process. See the &quot;Shutdown&quot; section for details. :id - Supervisor child id of the scheduler process. Defaults to Periodic. If you plan on running multiple periodic jobs under the same supervisor, make sure that they have different id values. :name - Registered name of the scheduler process. If not provided, the process will not be registered. :telemetry_id - Id used in telemetry event names. See the &quot;Telemetry&quot; section for more details. If not provided, telemetry events won&#39;t be emitted. :mode - When set to :manual, the jobs won&#39;t be started automatically. Instead you have to manually send tick signals to the scheduler. This should be used only in :test mix env. See the &quot;Testing&quot; section for details. Delay mode In the :regular mode (which is the default), the interval indicates time between two consecutive starts. This mode is typically useful if you want to maintain a stable execution rate (the number of executions per some time period). It is also a better choice if you&#39;re implementing fixed scheduling, as advised in the &quot;Fixed scheduling&quot; section. In the :shifted mode the interval represents the pause time between the end of the job and the start of the next one. This mode is likely a better choice if you want to have a fixed &quot;cool off&quot; period between two consecutive executions, to reduce the load on the system. Internally, Periodic relies on Erlang&#39;s monotonic time, which improves rate stability regardless of system time changes (see Time correction). Consider using the &quot;Multi-time warp mode&quot; (see here) to further improve rate stability in the situations when system time changes. In general, the overhead introduced by Periodic as well as job processing will be compensated, and you can usually expect stable intervals with very small variations (typically in sub milliseconds range), and no steady shift over time. However, in some cases, for example when the system is overloaded, the variations might be more significant. In the :shifted mode the job duration will affect the execution of the next job. In addition, Periodic will induce a slight (usually less than 100 microseconds), but a steady skew, due to its own internal processing. Shutdown To stop the scheduler, you need to ask its parent supervisor to stop the scheduler using Supervisor.terminate_child. The scheduler process acts as a supervisor, and so it has the same shutdown behaviour. When ordered to terminate by its parent, the scheduler will stop currently running job instances according to the :job_shutdown configuration. The default behaviour is to wait 5 seconds for the job to finish. However, in order for this waiting to actually happen, you need to invoke Process.flag(:trap_exit, true) from the run function. You can change the waiting time with the :job_shutdown option, which has the same semantics as in Supervisor. See [corresponding Supervisor documentation] (https://hexdocs.pm/elixir/Supervisor.html#module-shutdown-values-shutdown) for details. Fixed scheduling Periodic doesn&#39;t have explicit support for scheduling jobs at some particular time (e.g. every day at midnight). However, you can implement this on top of the provided functionality using the :when option defmodule SomeCleanup do def child_spec(_arg) do Periodic.child_spec( # check every minute if we need to run the cleanup every: :timer.minutes(1), # start the job only if it&#39;s midnight when: fn -&gt; match?(%Time{hour: 0, minute: 0}, Time.utc_now()) end, # ... ) end # ... end Note that the execution guarantees here are &quot;at most once&quot;. If the system is down at the scheduled time, the job won&#39;t be executed. Stronger guarantees can be obtained by basing the conditional logic on some persistence mechanism. Note that the :when guard is executed in the scheduler process. If the guard execution time is larger than the ticking period, time drifts will occur. Telemetry The scheduler optionally emits telemetry events. To configure telemetry you need to provide the :telemetry_id option. For example: Periodic.start_link(telemetry_id: :db_cleanup, ...) This will emit various events in the shape of [Periodic, telemetry_id, event]. Currently supported events are: :started - a new job instance is started :finished - job instance has finished or crashed (see related metadata for the reason) :skipped - new instance hasn&#39;t been started because the previous one is still running :stopped_previous - previous instance has been stopped because the new one is about to be started To consume the desired events, install the corresponding telemetry handler. Logging Basic logger is provided in Periodic.Logger. To use it, the scheduler needs to be started with the :telemetry_id option. To install logger handlers, you can invoke Periodic.Logger.install(telemetry_id). This function should be invoked only once per each scheduler during the system lifetime, preferably before the scheduler is started. A convenient place to do it is your application start callback. Testing The scheduler can be deterministically tested by setting the :mode option to :manual. In this mode, the scheduler won&#39;t tick on its own, and so it won&#39;t start any jobs unless instructed to by the client code. The :mode should be set to :manual only in test mix environment. Here&#39;s a simple approach which doesn&#39;t require app env and config files: defmodule MyPeriodicJob do @mode if Mix.env() != :test, do: :auto, else: :manual def child_spec(_arg) do Periodic.child_spec( mode: @mode, name: __MODULE__, # ... ) end # ... end Of course, you can alternatively use app env or any other approach you prefer. Just make sure to set the mode to manual only in test env. Notice that we&#39;re also setting the registered name and telemetry id. We&#39;ll need both to interact with the scheduler With such setup in place, the general shape of the periodic job test would look like this: def MyPeriodicJobTest do use ExUnit.Case, async: true require Periodic.Test test &quot;my periodic job&quot; do bring_the_system_into_the_desired_state() # tick the scheduler assert Periodic.Test.sync_tick(MyPeriodicJob) == {:ok, :normal} verify_side_effect_of_the_job() end end Note that this won&#39;t suffice for fixed schedules. Consider again the cleanup job which runs at midnight: defmodule SomeCleanup do def child_spec(_arg) do Periodic.child_spec( every: :timer.minutes(1), when: fn -&gt; match?(%Time{hour: 0, minute: 0}, Time.utc_now()) end, # ... ) end # ... end Manually ticking won&#39;t start the job, unless the test is running exactly at midnight. To make this module testable, you need to use a different implementation of :when in test environment: defmodule SomeCleanup do def child_spec(_arg) do Periodic.child_spec( every: :timer.minutes(1), when: &amp;should_run?/0 # ... ) end if Mix.env() != :test do defp should_run?(), do: match?(%Time{hour: 0, minute: 0}, Time.utc_now()) else defp should_run?(), do: true end # ... end Comparison to other schedulers There are various other abstractions for running periodic jobs in BEAM, such as: the built-in :timer module from Erlang stdlib erlcron quantum Oban Compared to :timer, Periodic offers some additional features, such as overlap handling, distributed scheduling, and telemetry support. Compared to most other third party libraries, Periodic will likely provide much less features out of the box. So in some situations, such as database persistence or back-pressure, you might need to invest more effort with Periodic. On the plus side Periodic should be simpler to use in typical scenarios, and much easier to reason about, while still providing enough flexibility to handle arbitrarily complex scenarios. For a more detailed discussion, see this blog post."},{"ref":"Periodic.html#child_spec/1","title":"Periodic.child_spec/1","type":"function","doc":"Builds a child specification for starting the periodic executor."},{"ref":"Periodic.html#start_link/1","title":"Periodic.start_link/1","type":"function","doc":"Starts the periodic executor."},{"ref":"Periodic.html#t:opts/0","title":"Periodic.opts/0","type":"type","doc":""},{"ref":"Periodic.Logger.html","title":"Periodic.Logger","type":"module","doc":"Telemetry handler that support basic logging of periodic scheduler events."},{"ref":"Periodic.Logger.html#install/1","title":"Periodic.Logger.install/1","type":"function","doc":"Installs telemetry handlers for the given scheduler."},{"ref":"Periodic.Test.html","title":"Periodic.Test","type":"module","doc":"Helpers for testing a periodic job. See the &quot;Testing&quot; section in Periodic documentation for details."},{"ref":"Periodic.Test.html#assert_periodic_event/4","title":"Periodic.Test.assert_periodic_event/4","type":"macro","doc":"Waits for the given telemetry event."},{"ref":"Periodic.Test.html#observe/1","title":"Periodic.Test.observe/1","type":"function","doc":"Subscribes to telemetry events of the given scheduler."},{"ref":"Periodic.Test.html#refute_periodic_event/4","title":"Periodic.Test.refute_periodic_event/4","type":"macro","doc":"Asserts that the given telemetry event won&#39;t be emitted."},{"ref":"Periodic.Test.html#sync_tick/2","title":"Periodic.Test.sync_tick/2","type":"function","doc":"Sends a tick signal to the given scheduler and waits for the job to finish. The function returns the job exit reason, or error if the job hasn&#39;t been started."},{"ref":"Periodic.Test.html#tick/1","title":"Periodic.Test.tick/1","type":"function","doc":"Sends a tick signal to the given scheduler. This function returns after the tick signal has been sent, and the job has been started. However, the function doesn&#39;t wait for the job to finish. If you want complete synchronism, use sync_tick/2"},{"ref":"rationale.html","title":"Rationale","type":"extras","doc":"Rationale When it comes to OTP supervision trees, the usual good practice is to run each process under some supervisor. In other words, a process which is a parent of other processes should most often be a supervisor. While this is most often a good approach, there are some cases where it&#39;s better to start processes directly under a GenServer or some other non-supervisor. Direct parenting requires the implementation to take some roles of the supervisor. The process should usually trap exits, keep track of its child processes, and ensure that all child processes are terminated before the parent itself stops. While this is not very hard to implement (and I advise doing it for practice), it will lead to some boilerplate. I&#39;ve personally encountered a couple of cases where I needed to code this from scratch, and it took some time and careful programming to get it right. The parent library is an attempt to extract the typical implementation into a common behaviour. Currently, the only provided behaviour is Parent.GenServer, which is basically a GenServer with built-in support for parenting. It&#39;s possible to create other behaviours (e.g. Parent.GenStage or Parent.GenStatem), with the help of the Parent module."},{"ref":"rationale.html#parent-genserver","title":"Rationale - Parent.GenServer","type":"extras","doc":"Parent.GenServer is a specialization of a GenServer. You can think of it as a mixture of GenServer, Supervisor, and Registry. The behaviour allows you to implement a server process which needs to parent one or more child processes, which can be started and stopped on-demand. The behaviour is used in a standard way: defmodule MyParentServer do use Parent.GenServer # ... end Once the behaviour is used, you can work with your module as if it was a GenServer. In other words you can use all the functions from the Elixir&#39;s GenServer module, and implement the same callbacks (e.g. init/1, handle_call/3, ...) with exactly the same interface and semantics. However, there are some differences: a server powered by Parent.GenServer by default traps exits. The behaviour will also take down all the known child processes on termination. Finally, the default version of injected child_spec/1 uses :infinity as the shutdown value, and :supervisor as the child type. The principal feature of Parent.GenServer is the ability to start managed child processes. This can be done with Parent.start_child/1, which must be invoked in the server process. The function takes a child specification as its single argument. The specification is a variation of a child spec map used with supervisors. In its full shape the child spec has the following type specification: %{ :id =&gt; term, start =&gt; {module, atom, [arg :: term]} | (() -&gt; {:ok, pid} | error :: term), optional(:shutdown) =&gt; non_neg_integer | :infinity | :brutal_kill, optional(:meta) =&gt; term } The shutdown field is optional, and if omitted, the default value of 5000 (5 seconds) is used. Just like with supervisors, a child spec can also be provided in the shape of {module, arg :: term} or module. The optional meta field can be used to associate arbitrary metadata to the child. Therefore, to start a task as a single child, you can write: defmodule MyParentServer do use Parent.GenServer def init(_) do {:ok, _child_pid} = Parent.start_child(%{ id: :hello_world, start: {Task, :start_link, [fn -&gt; IO.puts &quot;Hello, World!&quot; end]} }) {:ok, nil} end end And now, to start the parent server, you can invoke Parent.GenServer.start_link(MyParentServer, nil). Once a process is started, casts and calls are issued through Elixir&#39;s GenServer module. The Parent.GenServer module doesn&#39;t export any function which can be used to manipulate the parent server from the outside. So to reiterate: a Parent.GenServer is a GenServer. You might wonder what&#39;s the benefit of start_child/1 vs plain Task.start_link. The first benefit is that any child started through start_child/1 is known to the behaviour. Therefore, while terminating, the behaviour will make sure to take all the child processes down. This guarantees that if the parent is taken down, the child processes will not be left lingering. Another benefit is that you can use other Parent functions to work with child processes started with start_child/1. You can enumerate the child processes, fetch individual children by their id, and know the count of alive child processes. This is where Parent serves as a kind of a unique registry. Finally, if a child terminates, the behaviour will handle the corresponding :EXIT message, and convert it into the callback handle_child_terminated(child_name :: term, meta :: term, pid, reason :: term, state). This is the only callback specific to Parent.GenServer. This pretty much covers the Parent.GenServer behaviour. In my opinion, it&#39;s a relatively small variation of GenServer, which makes it fairly easy to reason about (assuming the user is already familiar with GenServer). However, before looking at use cases, let&#39;s discuss one controversial implementation detail. State management Careful readers might have noticed that the behaviour manages its state implicitly. When Parent.start_child/1 succeeds, the function returns {:ok, pid}, while the internal behaviour state is implicitly changed. In other words, start_child/1 is a function with a side-effect. The main reason why this approach is chosen is because starting of a child is by itself a side-effect operation. Thus, as soon as the new child is started, we want it to be included in other behaviour functions (e.g. when enumerating child processes or mapping a name to a pid). Using implicit state allows exactly this and leaves no place for mistakes. As an added bonus, the implicit state simplifies the interface of the callbacks, and allows Parent.GenServer to work exactly as a GenServer, which simplifies the transition to a parented version, and lowers the learning curve. For state storage, the behaviour uses the controversial process dictionary. This approach is chosen, since it keeps the entire state inside the process, so there&#39;s no data-copying involved nor are extra BEAM resources used (which would be the case if ETS or e.g. Agent was chosen)."},{"ref":"rationale.html#examples","title":"Rationale - Examples","type":"extras","doc":"Before looking at candidate examples, I want to reiterate that Parent.GenServer does not intend to replace supervisors. As mentioned in the introduction, in most cases you should run the workers directly under a supervisor. That said, I&#39;ve encountered some cases where it was better to directly parent some processes. In most of such cases, I&#39;ve noticed the following properties: The parent needs to detect termination of its child and perform some arbitrary logic when it happens. Children are one-off jobs which finish in a finite amount of time. Children are not used as servers for other processes in the system. However, in some cases they might be powered by server behaviours such as GenServer. The concrete examples I&#39;ve seen include periodic job execution, cancelable jobs with timeout, and job queues. Let&#39;s see how each example can be handled with the parent library. Periodic job execution The task at hand is to periodically (say every x seconds) run some job. It could be a cleanup job, communication with an external service, or anything else. Here&#39;s one way to implement this: Task.start_link(fn -&gt; fn -&gt; Process.sleep(:timer.seconds(5)) do_something() end |&gt; Stream.repeatedly() |&gt; Stream.run() end) This is definitely a simple solution which works out of the box. I occasionally use this approach and it works well in simple cases. However, there are some possible problems, depending on what do_something does. One issue is that if the job function crashes, the periodic job is also taken down. The supervisor will restart the process, but we need to be careful not to trip maximum restart intensity. In addition, this implementation has a problem of a possible drift. A long running computation might postpone the next job execution. If the computation gets stuck forever, do_something will not be executed anymore. Even crashes and restarts might cause some drift. Finally, if do_something allocates a lot of memory, it might put a pressure on the garbage collector, or lead to increased memory consumption. These issues can be solved if &quot;ticking&quot; is separated from the job execution. In other words, we&#39;d use one process, let&#39;s call it ticker, which start the job as a separate process every five seconds. If you want to follow the standard OTP way, where every worker is parented by a supervisor, you&#39;d likely need two supervisors here. Here&#39;s the supervision tree: TopSupervisor (one_for_all) / \\ JobSupervisor Ticker (GenServer) | Job (Task) The reason why we need TopSupervisor is to allow us to properly stop the periodic job execution without stopping anything else in the system. The reason why it&#39;s one_for_all is to prevent us from running multiple instances of the same job, if e.g. ticker is restarted. This is a fairly complex process structure for something as simple as &quot;invoke some function every X seconds&quot;. Moreover, these supervisors don&#39;t really lift much (if any) responsibilities from the ticker server. For example, if you want to prevent overlaps (running two instances of the same job at the same time), the ticker needs to keep track of currently running jobs, which means it will have to monitor them, and handle corresponding :DOWN messages. This will add some complexity to a server with a fairly simple basic role (start a job every X seconds). Therefore, I usually resort to starting a job as a direct child under the ticker process. This will simplify the supervision tree, but it will still require some complexity in the ticker server. The Parent.GenServer can help us here. Here&#39;s a naive implementation, which starts a job every second. The job generates a random number x in the range of 1-2000, then sleeps for x milliseconds, and prints the number. The ticker will not start another job if the previous one is still running: defmodule Demo.Periodic do use Parent.GenServer def start_link(), do: Parent.GenServer.start_link(__MODULE__, nil) @impl GenServer def init(_) do :timer.send_interval(:timer.seconds(1), :run_job) {:ok, nil} end @impl GenServer def handle_info(:run_job, state) do if Parent.child?(:job) do IO.puts(&quot;previous job already running&quot;) else Parent.start_child(%{ id: :job, start: {Task, :start_link, [&amp;job/0]} }) end {:noreply, state} end def handle_info(unknown_message, state), do: super(unknown_message, state) defp job() do num = :rand.uniform(2000) Process.sleep(num) IO.puts(num) end end"},{"ref":"rationale.html#cancellation-and-timeout","title":"Rationale - Cancellation and timeout","type":"extras","doc":"In this scenario we have the following needs: Start a job on user&#39;s request. Inform the user about sucess or failure. Stop the job if it has been running for too long. Stop the job on user&#39;s request. A simple sketch is given in the following code: defmodule Demo.Cancellable do use Parent.GenServer def start_link(), do: Parent.GenServer.start_link(__MODULE__, nil) def stop(pid), do: GenServer.stop(pid) @impl GenServer def init(_) do Parent.start_child(%{ id: :job, start: {Task, :start_link, [&amp;job/0]} }) Process.send_after(self(), :timeout, :timer.seconds(1)) {:ok, nil} end @impl GenServer def handle_info(:timeout, state) do IO.puts(&quot;timeout&quot;) {:stop, :timeout, state} end def handle_info(unknown_message, state), do: super(unknown_message, state) @impl Parent.GenServer def handle_child_terminated(:job, _meta, _pid, reason, state) do if reason == :normal do IO.puts(&quot;job succeeded&quot;) else IO.puts(&quot;job failed&quot;) end {:stop, reason, state} end defp job() do if :rand.uniform(5) == 1, do: raise(&quot;error&quot;) num = :rand.uniform(1500) Process.sleep(num) IO.puts(num) end end When you invoke Demo.Cancellable.start_link(), a job will be started. Depending on chance, the job will be succeed, fail (crash), or timeout. Each outcome is properly handled by its parent, and the failures are propagated further up the supervision tree. Moreover, you can cancel the job manually: iex&gt; {:ok, pid} = Demo.Cancellable.start_link(); Demo.Cancellable.stop(pid)"},{"ref":"rationale.html#simple-queue-with-limited-concurrency","title":"Rationale - Simple queue with limited concurrency","type":"extras","doc":"In this demo, we&#39;ll implement a job queue. The queue is a parent server which starts each job as a child task. The queue uses limited concurrency, with no more than 10 simultaneous jobs running at any point in time. Here&#39;s the full implementation: defmodule Demo.Queue do use Parent.GenServer def start_link() do Parent.GenServer.start_link(__MODULE__, nil, name: __MODULE__) end def add_jobs(count) do GenServer.call(__MODULE__, {:add_jobs, count}) end @impl GenServer def init(_) do {:ok, %{pending: :queue.new()}} end @impl GenServer def handle_call({:add_jobs, count}, _from, state) do state = Enum.reduce(1..count, state, &amp;push_item(&amp;2, &amp;1)) {:reply, :ok, state} end @impl Parent.GenServer def handle_child_terminated(_id, _meta, _pid, reason, state) do if reason == :normal do IO.puts(&quot;job succeeded&quot;) else IO.puts(&quot;job failed&quot;) end {:noreply, maybe_start_next_job(state)} end defp push_item(state, item) do state.pending |&gt; update_in(&amp;:queue.in(item, &amp;1)) |&gt; maybe_start_next_job() end defp maybe_start_next_job(state) do with max_jobs = 10, true &lt;- Parent.num_children() &lt; max_jobs, {{:value, _item}, remaining} &lt;- :queue.out(state.pending) do start_job() maybe_start_next_job(%{state | pending: remaining}) else _ -&gt; state end end defp start_job() do Parent.start_child(%{ id: make_ref(), start: {Task, :start_link, [&amp;job/0]} }) end defp job() do num = :rand.uniform(1500) Process.sleep(num) if :rand.uniform(5) == 1, do: raise(&quot;error&quot;) end end To use it, start the queue, and add some jobs: Demo.Queue.start_link() Demo.Queue.add_jobs(100) # enqueue 100 jobs You should see how some jobs succeed, while other fail. You might wonder how this compares to poolboy and similar pooling libraries. While there&#39;s some overlap, I believe that parent-based queue serves different cases. If you need to manage long-running resource oriented servers (such as db connections), you should opt for poolboy. On the other hand, I think that parent is more suitable for one-offs, since it runs every job in a separate process (which can admittedly be obtained with poolboy using max_overflow). In addition, the parent solution gives you a good amount of flexibility, allowing you for example to run jobs of different kinds, or implement more elaborate flow between jobs (e.g. don&#39;t start job bar before foo complets successfully, but allow baz to start unconditionally, which is in fact the scenario I&#39;ve had in practice)."},{"ref":"rationale.html#final-thoughts","title":"Rationale - Final thoughts","type":"extras","doc":"I&#39;ve previously implemented all of the scenarios above using vanilla GenServer, and I find the parent-based code to be much more focused on it&#39;s main job (periodic execution, cancellation, queuing). This is precisely the goal of the parent library. I&#39;ve seen such scenarios often enough to be annoyed by manual implementation from scratch. Beyond just saving typing and improving the reading experience, I think that the abstraction reduces the potential for accidental bugs. A manual implementation has to keep track of running child processes, worry about :EXIT or :DOWN messages, and make sure to terminate the child processes before it stops. All of this is automatically taken care of by the Parent.GenServer. Finally, compared to high-level out-of-the-box solutions such as quantum and poolboy, parent is less feature rich, but I think it&#39;s easier to reason about, and supports a wider set of scenarios. On the downside, this is in some ways a dangerous behaviour. It should be made clear that parent is not appropriate in all scenarios, and that standard supervision is most often the recommended approach. Since the parent process is not a supervisor, code reloading won&#39;t work (though I think we could make it work). That&#39;s why I feel that parent is suitable to parent only jobs, i.e. processes which stop in a finite amount of time, and don&#39;t serve requests from other processes in the system."}]